import inspect
from typing import Any, Generic, Self, TypeVar, final

import msgpack
from pydantic import TypeAdapter

from attp_client.misc.fixed_basemodel import FixedBaseModel

T = TypeVar("T")


class Serializable(Generic[T]):
    def __init__(self, data: T, enable_validation: bool = False) -> None:
        self.data = data
        self.enable_validation = enable_validation
    
        self.validate()
    
    def validate(self):
        if isinstance(self.data, FixedBaseModel) and self.enable_validation:
            self.data = self.data.model_dump(mode="json")
    
    @staticmethod
    def deserialize(obj: bytes, mp_configs: dict[str, Any] | None = None) -> Any:
        return msgpack.unpackb(obj, **(mp_configs or {}))
    
    def serialize(self, mp_configs: dict[str, Any] | None = None) -> bytes | None:
        return msgpack.packb(self.data, **(mp_configs or {}))
    
    @final
    @classmethod
    def mps(
        cls, 
        obj: bytes,
        mp_configs: dict[str, Any] | None = None
    ) -> Self:
        """
        Message Pack Serialize
        
        Deserializes and unpacks the model from the binary by utilizing Message Pack library.

        Opposite method: `mpd(...)`

        Parameters
        ----------
        obj : bytes
            Binary packed by Message Pack object.
        """
        return cls(data=cls.deserialize(obj, mp_configs=mp_configs))
    
    @final
    def mpd(self, mp_configs: dict[str, Any] | None = None) -> bytes | None:
        """
        Message Pack Dump
        
        Dumps and packs the model to the binary by utilizing Message Pack library.
        
        Opposite method: `mps(...)`
        """
        return self.serialize(mp_configs)
    
    def __setattr__(self, name: str, value: Any) -> None:
        caller = inspect.stack()[1].function
        if caller == "__init__" or caller.startswith("_"):
            super().__setattr__(name, value)
            return
        
        if name in ["data"]:
            raise TypeError(f"'Serializable' object does not support data mutation for attribute '{name}'")


from typing import TYPE_CHECKING, Any

import msgpack
from attp_client.core.utils.executor import execute_validated
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable
from attp_client.types.route_mapping import AttpRouteMapping
from attp_core.rs_api import PyAttpMessage, AttpCommand

from attp_client.utils.stream_object import StreamObject

if TYPE_CHECKING:
    from attp_client.session import SessionDriver


async def execute_call(
    frame: PyAttpMessage, 
    mapping: AttpRouteMapping,
    *, session: "SessionDriver"
):
    callback = mapping.callback
    
    payload = {}
    
    if frame.payload:
        payload = msgpack.unpackb(frame.payload)
    
    response = await execute_validated(callback, payload, frame=frame)
    
    if isinstance(response, StreamObject):
        iterable = response.iterate()
        if not iterable:
            return

        assert frame.correlation_id
        await session.respond_stream(
            route=frame.route_id,
            correlation_id=frame.correlation_id,
            command_type=AttpCommand.STREAMBOS,
        )

        if response.is_async:
            async for chunk in iterable:  # type: ignore
                await session.respond_stream(
                    route=frame.route_id,
                    correlation_id=frame.correlation_id,
                    payload=chunk,
                    command_type=AttpCommand.CHUNK,
                )
        else:
            for chunk in iterable:  # type: ignore
                await session.respond_stream(
                    route=frame.route_id,
                    correlation_id=frame.correlation_id,
                    payload=chunk,
                    command_type=AttpCommand.CHUNK,
                )

        await session.respond_stream(
            route=frame.route_id,
            correlation_id=frame.correlation_id,
            command_type=AttpCommand.STREAMEOS,
        )
        return

    if not isinstance(response, FixedBaseModel) and not isinstance(response, Serializable):
        response = Serializable[Any](data=response)
    
    assert frame.correlation_id
    await session.respond(route=frame.route_id, correlation_id=frame.correlation_id, payload=response)


async def execute_event(
    frame: PyAttpMessage,
    mapping: AttpRouteMapping
):
    callback = mapping.callback
    
    payload = {}
    
    if frame.payload:
        payload = msgpack.unpackb(frame.payload)
    
    await execute_validated(callback, payload, frame=frame)



from contextvars import ContextVar
from typing import Any, AsyncIterable, Callable, TypeVar, overload
import msgpack
from uuid import uuid4
from pydantic import TypeAdapter
from attp_core.rs_api import PyAttpMessage

from attp_client.core.ack_gate import StatefulAckGate
from attp_client.errors.dead_session import DeadSessionError
from attp_client.errors.serialization_error import SerializationError
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable
from attp_client.session import SessionDriver
from attp_client.utils.stream_receiver import StreamReceiver


T = TypeVar("T")
S = TypeVar("S")


class AttpRouter:
    def __init__(
        self,
        session: SessionDriver
    ) -> None:
        self.session = session
        self.context = ContextVar[str | None]("session_context", default=None)
        self.ack_gate = StatefulAckGate()
    
    def convert_message(self, expected_type: type[T], message: PyAttpMessage) -> T | Any:
        response = self.__format_response(expected_type=expected_type, response_data=message)
        return response
    
    @overload
    async def send(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 20,
    ) -> Any: ...
    
    @overload
    async def send(
        self, 
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 20, *,
        expected_response: type[T],
    ) -> T | Any: ...
    
    async def send(
        self, 
        route: str, 
        data: FixedBaseModel | Serializable | None = None,
        timeout: float = 50, *,
        expected_response: type[T] | None = None
    ) -> T | Any:
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)

        correlation_id = uuid4().bytes
        queue = await self.ack_gate.request_ack(correlation_id)
        try:
            await self.session.send_message(route=route, data=data, correlation_id=correlation_id)
            response_data = await self.ack_gate.wait_for_ack(correlation_id, timeout, queue=queue)
        finally:
            await self.ack_gate.complete_ack(correlation_id)
        
        return self.__format_response(expected_type=expected_response or Any, response_data=response_data)
    
    @overload
    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 50,
        *,
        formatter: Callable[[PyAttpMessage], S | None],
    ) -> AsyncIterable[S]: ...

    @overload
    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = ...,
        timeout: float = 50,
        *,
        formatter: None = ...,
    ) -> AsyncIterable[Any]: ...

    async def request_stream(
        self,
        route: str,
        data: FixedBaseModel | Serializable | None = None,
        timeout: float = 50,
        *,
        formatter: Callable[[PyAttpMessage], S | None] | None = None,
    ) -> AsyncIterable[Any] | AsyncIterable[S]:
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)

        correlation_id = uuid4().bytes
        queue = await self.ack_gate.request_ack(correlation_id)
        try:
            await self.session.send_message(route=route, data=data, correlation_id=correlation_id)
        except Exception:
            await self.ack_gate.complete_ack(correlation_id)
            raise

        async def _stream():
            try:
                async for frame in self.ack_gate.stream_ack(correlation_id, timeout, queue=queue):
                    yield frame
            finally:
                await self.ack_gate.complete_ack(correlation_id)

        stream = StreamReceiver(_stream(), formatter=formatter)
        
        return stream
        
    
    async def emit(self, route: str, data: FixedBaseModel | Serializable | None = None):
        if not self.session.is_connected:
            raise DeadSessionError(self.session.organization_id)
        
        await self.session.emit_message(route, data)

    async def handle_response(self, message: PyAttpMessage) -> None:
        if not message.correlation_id:
            return
        await self.ack_gate.feed(message)

    def __format_response(self, expected_type: Any, response_data: PyAttpMessage):
        if issubclass(expected_type, FixedBaseModel):
            if not response_data.payload:
                raise SerializationError(f"Nonetype payload received from session while expected type {expected_type.__name__}")
            try:
                return expected_type.mps(response_data.payload)
            except Exception as e:
                raise SerializationError(str(e))
        
        serialized = msgpack.unpackb(response_data.payload) if response_data.payload else None
        
        if expected_type is not None:
            return serialized
        
        return TypeAdapter(expected_type, config={"arbitrary_types_allowed": True}).validate_python(serialized)



from typing import Any, Mapping

from attp_client.interfaces.error import IErr


class AttpException(Exception):
    def __init__(self, code: str = "UnknownError", *, detail: Mapping[str, Any]) -> None:
        self.code = code
        self.detail = detail
    
    def to_ierr(self):
        return IErr(detail={"code": self.code, **self.detail})
    
    @staticmethod
    def from_ierr(err: IErr, **kwargs):
        code = err.detail.pop("code", "UnknownError")
        
        return AttpException(code=code, detail={**err.detail, **kwargs})


import asyncio
from logging import Logger, getLogger
from typing import Any
from attp_core.rs_api import AttpClientSession, Limits
from attp_core.rs_api import PyAttpMessage

from attp_client.catalog import AttpCatalog
from attp_client.inference import AttpInferenceAPI
from attp_client.interfaces.catalogs.catalog import ICatalogResponse
from attp_client.misc.serializable import Serializable
from attp_client.objects.agents import AttpAgents
from attp_client.objects.chats import AttpChats
from attp_client.core.dispatcher import AttpDispatcher
from attp_client.core.eventbus import AttpEventBus
from attp_client.core.receiver import AttpReceiver
from attp_client.core.tool_callbacks import ToolCallbacks
from attp_client.router import AttpRouter
from attp_client.session import SessionDriver
from attp_client.tools import ToolsManager

from attp_core.rs_api import init_logging


class ATTPClient:
    
    is_connected: bool
    client: AttpClientSession
    session: SessionDriver | None
    inference: AttpInferenceAPI
    catalogs: list[AttpCatalog]

    _tools: ToolsManager | None
    _tool_callbacks: ToolCallbacks | None
    eventbus: AttpEventBus | None
    receiver: AttpReceiver[tuple[SessionDriver, PyAttpMessage]] | None
    dispatcher: AttpDispatcher | None

    def __init__(
        self,
        agt_token: str,
        organization_id: int,
        *,
        connection_url: str | None = None,
        reconnect: bool = False,
        max_retries: int = 20,
        limits: Limits | None = None,
        logger: Logger | None = None,
        verbose: bool = False
    ):
        self.__agt_token = agt_token
        self.organization_id = organization_id
        self.connection_url = connection_url or "attp://localhost:6563"
        
        self.session = None
        self.max_retries = max_retries
        self.limits = limits or Limits(max_payload_size=50000)
        self.client = AttpClientSession(self.connection_url, limits=self.limits)
        self.logger = logger or getLogger("Ascender Framework")
        self.reconnect = reconnect
        
        self.catalogs = []
        self._client = None
        self.verbose = verbose
        
        self._tools = None
        self._tool_callbacks = None
        self.eventbus = None
        self.receiver = None
        self.dispatcher = None
        
        if self.verbose:
            init_logging()
    
    async def connect(self):
        # Open the connection
        client = await self.client.connect(self.max_retries)
        self._client = client
        
        if not client.session:
            raise ConnectionError("Failed to connect to ATTP server after 10 attempts!")
        
        self.session = SessionDriver(
            client.session, 
            agt_token=self.__agt_token, 
            organization_id=self.organization_id,
            factory=self.client,
            on_reconnect=self._reconnect if self.reconnect else None,
            logger=self.logger or getLogger("Ascender Framework")
        )

        self.router = AttpRouter(self.session)
        self.eventbus = AttpEventBus(self.router)
        self.receiver = AttpReceiver()
        self.dispatcher = AttpDispatcher(self.eventbus, self.router)
        self.dispatcher.start(self.receiver)
        self.session.set_receiver(self.receiver)

        self._tools = ToolsManager(self.router)
        self._tool_callbacks = ToolCallbacks(
            session=self.session,
            catalogs=self.catalogs,
            tools=self._tools,
        )
        await self._tool_callbacks.register(self.eventbus)
        
        asyncio.create_task(self.session.start_listener())
        await self.session.authenticate(list(self.eventbus.routes))
        self.inference = AttpInferenceAPI(self.router)
        self.chats = AttpChats(self.router)
        self.agents = AttpAgents(self.router)
    
    async def _reconnect(self):
        self.logger.info("Attempting to reconnect to ATTP server...")
        await self.connect()

    async def close(self):
        if self.session:
            self.session.on_reconnect = None

            if self.dispatcher:
                self.dispatcher.stop()

            await self.session.close()
            
            self.session = None
            self.is_connected = False

    @property
    def tools(self):
        if not self._tools:
            self._tools = ToolsManager(self.router)
        
        return self._tools
    
    async def catalog(self, catalog_name: str):
        if any(c.catalog_name == catalog_name for c in self.catalogs):
            return next(c for c in self.catalogs if c.catalog_name == catalog_name)
        
        catalog = await self.router.send(
            "tools:catalogs:specific", 
            Serializable[dict[str, str]]({"catalog_name": catalog_name}),
            timeout=10,
            expected_response=ICatalogResponse
        )
        self.catalogs.append(
            AttpCatalog(id=catalog.catalog_id, catalog_name=catalog_name, manager=self.tools)
        )

        return self.catalogs[-1] # Return the newly added catalog

    async def close_catalog(self, catalog: AttpCatalog):
        catalog.detach_all_tools()
        self.catalogs.remove(catalog)



import asyncio
import traceback
from logging import Logger, getLogger
from typing import Any, Awaitable, Callable, Sequence
from uuid import uuid4


from attp_client.consts import ATTP_VERSION
from attp_client.core.receiver import AttpReceiver
from attp_client.errors.dead_session import DeadSessionError
from attp_client.errors.unauthenticated_error import UnauthenticatedError
from attp_client.interfaces.handshake.auth import IAuth
from attp_client.interfaces.error import IErr
from attp_client.interfaces.handshake.hello import IHello
from attp_client.interfaces.handshake.ready import IReady
from attp_client.interfaces.route_mappings import IRouteMapping
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_core.rs_api import Session, PyAttpMessage, AttpCommand, AttpClientSession

from attp_client.misc.serializable import Serializable
from attp_client.types.route_mapping import AttpRouteMapping
from attp_client.utils import serializer
from attp_client.utils.route_mapper import resolve_route_by_id



class SessionDriver:
    _organization_id: int
    server_routes: Sequence[IRouteMapping] | None
    receiver: AttpReceiver[tuple["SessionDriver", PyAttpMessage]] | None
    
    def __init__(
        self, 
        session: Session, 
        agt_token: str,
        organization_id: int,
        *,
        factory: AttpClientSession,
        on_reconnect: Callable[[], Any] | None = None,
        # route_mappings: Sequence[AttpRouteMapping],
        logger: Logger = getLogger("Ascender Framework"),
    ) -> None:
        self.agt_token = agt_token
        self.on_reconnect = on_reconnect
        self.factory = factory
        self.session = session
        self._organization_id = organization_id
        self.server_routes = None
        self.logger = logger
        
        self.client_routes = []
        # self.is_authenticated = False
        self.receiver = None
        
        self.messages = asyncio.Queue[PyAttpMessage]()
        self.auth_event = asyncio.Event()
    
    @property
    def is_connected(self) -> bool:
        return bool(self.session)
    
    @property
    def session_id(self) -> str | None:
        if not self.session:
            raise DeadSessionError(self.organization_id)
        
        return self.session.session_id

    @property
    def peername(self) -> str:
        if not self.session:
            return "undefined"
        
        return self.session.peername or "undefined"
    
    @property
    def is_authenticated(self) -> bool:
        return self.auth_event.is_set()
    
    @property
    def organization_id(self) -> int:
        return self._organization_id
    
    async def send_raw(self, frame: PyAttpMessage):
        """
        Send raw message to session driver.

        Parameters
        ----------
        frame : PyAttpMessage
            Attp Frame that contains 
        """
        if not self.session:
            raise DeadSessionError(self.organization_id)
        
        return await self.session.send(frame)

    def set_receiver(self, receiver: AttpReceiver[tuple["SessionDriver", PyAttpMessage]] | None) -> None:
        self.receiver = receiver
    
    async def send_message(
        self,
        route: str | int,
        data: FixedBaseModel | Serializable | None,
        *,
        correlation_id: bytes | None = None
    ) -> bytes:
        """
        Sends an ATTPMessage to the client.

        Parameters
        ----------
        route : str | int
            String pattern of the route if str passed, or int ID of route.
        data : FixedBaseModel | Serializable | None
            A serializable data that will be sent.
        correlation_id : bytes | None, optional
            Optional correlation ID to use for the request.
        
        Returns
        -------
        bytes
            Correlation ID used for mapping the response.
        """
        if not self.session:
            raise DeadSessionError(self.organization_id)
        
        if not self.server_routes:
            raise UnauthenticatedError(f"Cannot send an ATTP message with acknowledgement to unauthenticated (route_mapping={route})")
        
        if correlation_id is None:
            correlation_id = uuid4().bytes
        relevant_route = route
        
        if isinstance(route, str):
            relevant_route = resolve_route_by_id("message", route, self.server_routes).route_id
        
        # print("RELEVANT ROUTE", relevant_route)
        
        frame = PyAttpMessage(int(relevant_route), AttpCommand.CALL, correlation_id=correlation_id, payload=data.mpd() if data is not None else None, version=ATTP_VERSION)
        # print(frame.payload)
        await self.send_raw(frame)
        
        return correlation_id
    
    
    async def emit_message(self, route: str | int, data: FixedBaseModel | Serializable | None) -> None:
        """
        Emits an ATTPMessage to the client.
        It forms the EMIT frame instead of CALL, which doesn't require receiver to respond to it.

        Parameters
        ----------
        route : str | int
            String pattern of the route if str passed, or int ID of route.
        data : FixedBaseModel | Serializable | None
            A serializable data that will be sent.
        """
        if not self.server_routes:
            raise UnauthenticatedError(f"Cannot send an ATTP message with acknowledgement to unauthenticated (route_mapping={route})")
        
        relevant_route = route
        
        if isinstance(route, str):
            relevant_route = resolve_route_by_id("event", route, self.server_routes).route_id
        
        frame = PyAttpMessage(int(relevant_route), AttpCommand.EMIT, correlation_id=None, payload=data.mpd() if data is not None else None, version=ATTP_VERSION)
        await self.send_raw(frame)
    
    async def authenticate(self, route_mappings: Sequence[AttpRouteMapping] | None) -> None:
        """
        Send AUTH frame to attp server (AgentHub).
        Version should be b'01' in bytes.

        Parameters
        ----------
        version : bytes
            Version of ATTP protocol (to validate on Rust side). It's 01, which correlated to '0.1'

        Returns
        -------
        tuple[str, int]
            (session_id, organization_id) - client sends organization_id while session ID is generated by Rust.
        """
        if route_mappings:
            self.client_routes.extend([IRouteMapping.from_route_mapper(mapper) for mapper in route_mappings])
        
        frame = PyAttpMessage(
            route_id=0, 
            command_type=AttpCommand.AUTH, 
            correlation_id=None,
            payload=IAuth(
                token=self.agt_token, 
                organization_id=self.organization_id
            ).mpd(),
            version=ATTP_VERSION
        )
        await self.send_raw(frame)
        
        await asyncio.wait_for(self.auth_event.wait(), 10)

    async def send_error(self, err: IErr, correlation_id: bytes | None = None, route: str | int = 0) -> None:
        """
        Send an error to the session peer.
        Can be two types, non correlated to ack and correlated to ack.
        
        When correlated to ack, the `correlation_id` as a response to acknowledgement is required to specify.

        Parameters
        ----------
        err : IErr
            Error details.
        correlation_id : bytes | None, optional
            For correlated to ack, ID of the correlation as a response error, by default None (non-correlated)
        route : str | int, optional
            For correlated to ack, not required and not allowed to specify route, but for non-correlated it's optional, by default 0
        """
        relevant_route = route
        
        if not self.server_routes:
            raise UnauthenticatedError(f"Cannot send an ATTP message with acknowledgement to unauthenticated (route_mapping={route})")
        
        
        if isinstance(route, str):
            relevant_route = resolve_route_by_id("err", route, self.server_routes).route_id
        
        await self.send_raw(
            PyAttpMessage(
                route_id=int(relevant_route),
                command_type=AttpCommand.ERR,
                correlation_id=correlation_id,
                payload=err.mpd(),
                version=ATTP_VERSION
            )
        )
    
    async def respond(self, route: int | str, correlation_id: bytes, payload: FixedBaseModel | Serializable | Any | None = None):
        """
        For responding to `AttpCommand.CALL`. Used only for correlated requests.
        It sends response (acknowledgement) message signed as `AttpCommand.ACK` to the request.

        Parameters
        ----------
        correlation_id : bytes
            Correlation ID to which response is being sent.
        payload : FixedBaseModel | Serializable | None, optional
            Response payload, the data that will be sent, by default None
        """
        relevant_route = route

        if not self.server_routes:
            raise UnauthenticatedError(f"Cannot send an ATTP message with acknowledgement to unauthenticated (route_mapping={route})")
        
        if isinstance(route, str):
            relevant_route = resolve_route_by_id("message", route, self.server_routes).route_id

        frame = PyAttpMessage(
            route_id=int(relevant_route),
            command_type=AttpCommand.ACK,
            correlation_id=correlation_id,
            payload=serializer.deserialize(payload),
            version=ATTP_VERSION
        )
        
        await self.send_raw(frame)

    async def respond_stream(
        self,
        route: int | str,
        correlation_id: bytes,
        *,
        payload: FixedBaseModel | Serializable | Any | None = None,
        command_type: AttpCommand,
    ):
        """
        Send a streaming response frame (STREAMBOS/CHUNK/STREAMEOS) for a correlated request.
        """
        relevant_route = route

        if not self.server_routes:
            raise UnauthenticatedError(
                f"Cannot send an ATTP stream message to unauthenticated session (route_mapping={route})"
            )

        if isinstance(route, str):
            relevant_route = resolve_route_by_id("message", route, self.server_routes).route_id

        frame = PyAttpMessage(
            route_id=int(relevant_route),
            command_type=command_type,
            correlation_id=correlation_id,
            payload=serializer.deserialize(payload),
            version=ATTP_VERSION,
        )

        await self.send_raw(frame)

    async def listen(self, on_message: Callable[[PyAttpMessage], Awaitable[None]]) -> None:
        """
        Start a background read-loop task that:
          - Routes CALL, EMIT -> `events` (apply backpressure with await put)
          - Routes ACK, ERR   -> response handler
        Must:
          - batch across the FFI boundary
          - enforce correlation rules:
              * CALL must include Correlation-Id
              * ACK/ERR must include Correlation-Id
          - propagate terminal errors by finishing the Task with an exception
          - stop cleanly on orderly close
        """
        while self.is_authenticated:
            message = await self.messages.get()
            await on_message(message)

    async def close(self):
        """
        Closes the connection from server-side between the client.
        """
        await self.send_raw(PyAttpMessage(
            route_id=0,
            command_type=AttpCommand.DISCONNECT,
            correlation_id=None,
            payload=None,
            version=ATTP_VERSION
        ))
        self.session.stop_listener()
        self.session.disconnect()
    
    async def handle_ready(self, frame: IReady):
        self.server_routes = frame.server_routes
        
        # print(frame.server_routes)
        
        data = PyAttpMessage(
            route_id=0,
            command_type=AttpCommand.READY,
            correlation_id=None,
            payload=IHello(proto="attp", ver="0.1", caps=[], mapping=self.client_routes).mpd(),
            version=b"01"
        )
        await self.send_raw(data)
        self.auth_event.set()
    
    async def _on_event(self, events: list[PyAttpMessage]) -> None:
        try:
            try:
                session_id = self.session_id
            except Exception:
                session_id = "unknown"

            self.logger.debug(f"[cyan]ATTP[/] ┆ Received a new message from session {session_id} ")
            # assert self.session
            
            for event in events:
                try:
                    if event.route_id == 0 and event.command_type == AttpCommand.READY:
                        if not event.payload:
                            continue
                        
                        try:
                            await self.handle_ready(IReady.mps(event.payload))
                        except Exception:
                            traceback.print_exc()
                            self.logger.exception("[cyan]ATTP[/] ┆ Failed to handle READY frame.")
                            await self.close()
                            break
                        continue
                    
                    if event.command_type == AttpCommand.DISCONNECT:
                        self.messages.put_nowait(event)
                        self.logger.info(f"[cyan]ATTP[/] ┆ Session {self.session_id} disconnected by the server.")
                        await self.close()
                        
                        if self.on_reconnect:
                            self.logger.info(f"[cyan]ATTP[/] ┆ Attempting to reconnect session for organization {self.organization_id}...")
                            try:
                                await self.on_reconnect()
                                self.logger.info(f"[cyan]ATTP[/] ┆ Successfully reconnected session for organization {self.organization_id}.")
                            except Exception:
                                traceback.print_exc()
                                self.logger.exception(f"[cyan]ATTP[/] ┆ Failed to reconnect session for organization {self.organization_id}.")
                    else:
                        if self.is_authenticated:
                            self.logger.debug("[cyan]ATTP[/] ┆ Handing incoming message to a route handler.")
                            if self.receiver:
                                self.receiver.on_next((self, event))
                            else:
                                self.messages.put_nowait(event)
                except Exception:
                    traceback.print_exc()
                    self.logger.exception(
                        "[cyan]ATTP[/] ┆ Failed to process incoming frame "
                        f"(route_id={event.route_id}, command_type={event.command_type})."
                    )
        except Exception:
            traceback.print_exc()
            self.logger.exception("[cyan]ATTP[/] ┆ Unhandled error in _on_event handler.")

    async def start_listener(self):
        if not self.session:
            raise DeadSessionError(self.organization_id)
        self.session.add_event_handler(self._on_event)
        
        await asyncio.gather(
            self.session.start_handler(),
            self.session.start_listener()
        )


from typing import Any, Literal
from uuid import UUID
from attp_client.interfaces.inference.enums.message_emergency_type import MessageEmergencyTypeEnum
from attp_client.interfaces.inference.enums.message_type import MessageTypeEnum
from attp_client.interfaces.inference.tool import ToolV2
from attp_client.misc.fixed_basemodel import FixedBaseModel


class BaseMessage(FixedBaseModel):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)

class IAttachmentDTO(BaseMessage):
    file_id: int

class IMessageDTOV2(FixedBaseModel):
    """
    The DTO Object of Message in AgentHub, contains message content and tool execution data.
    """
    content: dict | str | None
    message_type: MessageTypeEnum
    attachments: list[IAttachmentDTO] | None = None

    reply_to_message_id: int | None = None
    chat_id: UUID
    agent_id: int | None = None
    user_id: int | None = None
    client_id: str | None = None

    tool_called: ToolV2 | None = None
    tool_status: Literal['started', 'finished', 'error'] | None = None
    tool_started_input: str | None = None
    tool_finished_output: str | None = None
    tool_error_detail: str | None = None

    specialist_required: MessageEmergencyTypeEnum | None = None
    meta: dict[str, Any] | None = None # key, value (key is string, value is JSON string, None as value should be 'null' in JSON)

    def to_wrap(self) -> dict:
        w = self.model_dump()
        del w['attachments']

        return w


class IMessageResponse(IMessageDTOV2):
    """
    Response model of AgentHub Message, it represents [MessageDTO](agenthub/dtos/message_dto.py).
    But it contains additional value as ID
    """
    id: int
    """ID of Message in database"""


from enum import Enum


class ChatModeEnum(Enum):
    AGENT_AUTOPILOT = "agent_autopilot"
    AGENT_CO_PILOT = "agent_copilot"
    PRIVATE_SINGLE = "private_single"
    PRIVATE_GROUP = "private_group"


from attp_client.types.route_mapping import AttpRouteMapping, RouteType
from attp_client.misc.fixed_basemodel import FixedBaseModel


class IRouteMapping(FixedBaseModel):
    pattern: str
    route_id: int
    route_type: RouteType
    
    @staticmethod
    def from_route_mapper(mapper: AttpRouteMapping):
        return IRouteMapping(pattern=mapper.pattern, route_id=mapper.route_id, route_type=mapper.route_type)


import asyncio
import traceback
from attp_client.core.eventbus import AttpEventBus
from attp_client.core.receiver import AttpReceiver
from attp_client.interfaces.error import IErr
from attp_client.router import AttpRouter
from attp_client.session import SessionDriver

from attp_core.rs_api import PyAttpMessage, AttpCommand


class AttpDispatcher:
    def __init__(self, eventbus: AttpEventBus, router: AttpRouter) -> None:
        self.eventbus = eventbus
        self.router = router
        self._task: asyncio.Task | None = None
        
    def start(self, receiver: AttpReceiver[tuple[SessionDriver, PyAttpMessage]]) -> None:
        if self._task and not self._task.done():
            return
        self._task = asyncio.create_task(self._run(receiver))

    def stop(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()

    async def _run(self, receiver: AttpReceiver[tuple[SessionDriver, PyAttpMessage]]) -> None:
        while True:
            session, msg = await receiver.get()
            try:
                if msg.command_type in (
                    AttpCommand.ACK,
                    AttpCommand.ERR,
                    AttpCommand.DEFER,
                    AttpCommand.STREAMBOS,
                    AttpCommand.CHUNK,
                    AttpCommand.STREAMEOS,
                ):
                    await self.router.handle_response(msg)
                else:
                    await self.eventbus.emit(session, msg)
            except Exception:
                traceback.print_exc()
                if msg.command_type == AttpCommand.CALL and msg.correlation_id:
                    await session.send_error(
                        IErr(detail={"code": "InternalDispatchError", "message": "Dispatcher failed to process frame."}),
                        correlation_id=msg.correlation_id,
                        route=msg.route_id
                    )
            finally:
                receiver.task_done()



from typing import Any

from attp_client.misc.fixed_basemodel import FixedBaseModel


class IAgentDTO(FixedBaseModel):
    """
    A shared DTO (Data Transfer Object) for Agents.
    """
    name: str
    avatar_url: str | None = None
    description: str
    module_id: str
    configurations: dict[str, Any] = {}


class IAgentResponse(IAgentDTO):
    id: int



from typing import Annotated
from pydantic import Field
from typing_extensions import Doc
from attp_client.misc.fixed_basemodel import FixedBaseModel


class IAuth(FixedBaseModel):
    token: Annotated[str | None, Doc("Provided AGT token for authentication")] = Field(None)
    organization_id: Annotated[int, Doc("ID of organization.")]


from typing import Any
from uuid import UUID

from attp_client.interfaces.error import IErr

class CorrelatedRPCException(Exception):
    correlation_id: UUID
    def __init__(
        self, 
        correlation_id: bytes,
        detail: dict[str, Any] | None = None,
    ):
        self.correlation_id = UUID(bytes=correlation_id)
        self.detail = detail
    
    @staticmethod
    def from_err_object(correlation_id: bytes, err: IErr):
        return CorrelatedRPCException(correlation_id, err.detail)
    
    def __str__(self) -> str:
        return f"[ATTPRPC ERR]: {self.detail}"


import asyncio
from typing import Generic, TypeVar


T = TypeVar("T")


class AttpReceiver(Generic[T]):
    def __init__(self) -> None:
        self._queue: asyncio.Queue[T] = asyncio.Queue()

    def on_next(self, item: T) -> None:
        self._queue.put_nowait(item)

    async def get(self) -> T:
        return await self._queue.get()

    def task_done(self) -> None:
        self._queue.task_done()



from attp_client.misc.fixed_basemodel import FixedBaseModel


class ICatalogResponse(FixedBaseModel):
    catalog_id: int
    organization_id: int


class ICatalogNameResponse(FixedBaseModel):
    catalog_name: str


import traceback
from typing import TYPE_CHECKING, Any, Callable, Set

from pydantic import ValidationError, validate_call

from attp_client.core.utils.callbacks import execute_call, execute_event
from attp_client.errors.attp_exception import AttpException
from attp_client.interfaces.error import IErr
from attp_client.session import SessionDriver
from attp_client.types.route_mapping import AttpRouteMapping, RouteType

if TYPE_CHECKING:
    from attp_client.router import AttpRouter

from attp_core.rs_api import PyAttpMessage, AttpCommand


class AttpEventBus:
    routes: Set[AttpRouteMapping]
    
    def __init__(self, router: "AttpRouter") -> None:
        self.routes = set()
        self.increment_index = 2
        self.router = router
    
    async def subscribe(
        self,
        route_type: RouteType,
        pattern: str,
        callback: Callable[..., Any]
    ):
        if pattern == "connect" and route_type == "connect":
            self.routes.add(AttpRouteMapping(
                pattern=pattern,
                route_id=0,
                route_type=route_type,
                callback=validate_call(config={"arbitrary_types_allowed": True})(callback)
            ))
            return
        
        if pattern == "disconnect" and route_type == "disconnect":
            self.routes.add(AttpRouteMapping(
                pattern=pattern,
                route_id=0,
                route_type=route_type,
                callback=validate_call(config={"arbitrary_types_allowed": True})(callback)
            ))
            return
        
        self.routes.add(AttpRouteMapping(
            pattern=pattern,
            route_id=self.increment_index,
            route_type=route_type,
            callback=validate_call(config={"arbitrary_types_allowed": True})(callback)
        ))
        self.increment_index += 1
    
    async def emit(self, session: SessionDriver, message: PyAttpMessage):
        if message.command_type not in (AttpCommand.EMIT, AttpCommand.CALL, AttpCommand.ERR):
            return
        
        try:   
            relevant_route = next(
                (route for route in self.routes if route.route_id == message.route_id),
                None
            )
            
            if message.command_type == AttpCommand.CALL:
                if not relevant_route:
                    assert message.correlation_id
                    await self.router.session.send_error(
                        IErr(detail={"code": "NotFound", "message": "Route not found."}), 
                        message.correlation_id, 
                        message.route_id or 0
                    )
                    return
                await execute_call(message, relevant_route, session=self.router.session)
            
            if message.command_type == AttpCommand.EMIT:
                if not relevant_route:
                    return
                
                await execute_event(message, relevant_route)
            # TODO: Implement error callback
            # Conditions to fire the callback, if err callback has specific route_id, fire it if frame has same route_id, if it doesn't (route_id is 1) then fire all relevant_routes.
            # If `message` frame has correlation_id then we should skip if (if route_ids won't match)
        except ValidationError as e:
            traceback.print_exc()
            await session.send_error(IErr(detail={"code": "ValidationError", "data": str(e)}), correlation_id=message.correlation_id, route=message.route_id)
        
        except AttpException as e:
            traceback.print_exc()
            await session.send_error(e.to_ierr(), correlation_id=message.correlation_id, route=message.route_id)
        
        except Exception:
            traceback.print_exc()
            await session.send_error(IErr(detail={"code": "InternalExecutionError", "data": traceback.format_exc()}), correlation_id=message.correlation_id, route=message.route_id)


import msgpack

from typing import AsyncIterable, AsyncIterator, Callable, Generic, TypeVar
from attp_core.rs_api import PyAttpMessage


T = TypeVar("T")


class StreamReceiver(Generic[T]):
    def __init__(
        self,
        _generator: AsyncIterable[PyAttpMessage],
        *,
        formatter: Callable[[PyAttpMessage], T | None] | None = None,
    ) -> None:
        self.generator = _generator
        self.formatter = formatter

    def __aiter__(self) -> AsyncIterator[T]:
        return self.__iter_stream()

    def default_formatter(self, data: PyAttpMessage):
        if not data.payload:
            return None
        
        return msgpack.unpackb(data.payload)

    async def __iter_stream(self):
        async for frame in self.generator:
            if not self.formatter:
                formatted = self.default_formatter(frame)
                if formatted is not None:
                    yield formatted
                continue

            formatted = self.formatter(frame)
            if formatted is not None:
                yield formatted



from typing import Any, Callable, MutableMapping
from uuid import UUID

from attp_client.errors.attp_exception import AttpException
from attp_client.errors.not_found import NotFoundError
from attp_client.interfaces.catalogs.tools.envelope import IEnvelope
from attp_client.tools import ToolsManager


class AttpCatalog:
    attached_tools: MutableMapping[str, Callable[..., Any]] # id, callback
    tool_name_to_id_symlink: MutableMapping[str, str] # name, id
    
    def __init__(
        self,
        id: int,
        catalog_name: str,
        manager: ToolsManager
    ) -> None:
        self.id = id
        self.catalog_name = catalog_name
        self.tool_manager = manager
        self.attached_tools = {}
        self.tool_name_to_id_symlink = {}
    
    async def handle_callback(self, envelope: IEnvelope) -> Any:
        if envelope.tool_id not in self.attached_tools:
            raise NotFoundError(f"Tool {envelope.tool_id} not marked as registered and wasn't found in the catalog {self.catalog_name}.")

        try:
            result = await self.handle_call(envelope)
        except AttpException as e:
            return e.to_ierr()
        
        return result

    def attach_tool(
        self,
        callback: Callable[[IEnvelope], Any],
        name: str, 
        description: str | None = None,
        schema: dict | None = None,
        schema_id: str | None = None,
        *,
        return_direct: bool = False,
        schema_ver: str = "1.0",
        timeout_ms: float = 20000,
        idempotent: bool = False
    ):
        assigned_id = self.tool_manager.register(
            self.catalog_name,
            name=name,
            description=description,
            schema_id=schema_id,
            return_direct=return_direct,
            schema=schema,
            schema_ver=schema_ver,
            timeout_ms=timeout_ms,
            idempotent=idempotent
        )
        
        self.attached_tools[str(assigned_id)] = callback
        self.tool_name_to_id_symlink[name] = str(assigned_id)
        return assigned_id
    
    def detach_tool(
        self,
        name: str
    ):
        tool_id = self.tool_name_to_id_symlink.get(name)
        
        if not tool_id:
            raise NotFoundError(f"Tool {name} not marked as registered and wasn't found in the catalog {self.catalog_name}.")
        
        self.tool_manager.unregister(self.catalog_name, UUID(tool_id))
        return tool_id
    
    def detach_all_tools(self):
        for tool_id in list(self.attached_tools.keys()):
            self.tool_manager.unregister(self.catalog_name, UUID(tool_id))
            del self.attached_tools[tool_id]
        
        self.tool_name_to_id_symlink.clear()
    
    async def handle_call(self, envelope: IEnvelope) -> Any:
        tool = self.attached_tools.get(envelope.tool_id)

        if not tool:
            raise NotFoundError(f"Tool {envelope.tool_id} not marked as registered and wasn't found in the catalog {self.catalog_name}.")

        return await tool(envelope)



import inspect
from typing import Any, Mapping, get_args, get_origin

from pydantic import BaseModel, TypeAdapter

from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_core.rs_api import PyAttpMessage


async def execute_validated(callback: Any, payload: Any, *, frame: PyAttpMessage | None = None):
    """Thx GPT-5 for the call validator!"""
    sig = inspect.signature(callback)
    params = list(sig.parameters.values())

    # If it's a bound method, drop the first "self"
    if inspect.ismethod(callback) and callback.__self__ is not None:
        params = params[1:]
        sig = sig.replace(parameters=params)

    def wants_frame(param: inspect.Parameter) -> bool:
        if frame is None:
            return False

        ann = param.annotation
        if ann is PyAttpMessage:
            return True

        origin = get_origin(ann)
        if origin is not None and PyAttpMessage in get_args(ann):
            return True

        return ann is inspect._empty and param.name in ("message", "frame")

    # --- Case 1: single-param message frame ---
    if len(params) == 1:
        param = params[0]
        if wants_frame(param):
            if inspect.iscoroutinefunction(callback):
                return await callback(frame)
            return callback(frame)

        # --- Case 2: single-param model ---
        ann = param.annotation

        if ann is not inspect._empty and (issubclass_safe(ann, FixedBaseModel) or issubclass_safe(ann, BaseModel)):
            model = ann.model_validate(payload) if hasattr(ann, "model_validate") else ann(**payload)
            if inspect.iscoroutinefunction(callback):
                return await callback(model)
            return callback(model)

    # --- Case 3: normal kwargs mapping ---
    bound_args = {}
    for name, param in sig.parameters.items():
        if name not in payload and wants_frame(param):
            bound_args[name] = frame
            continue

        if name not in payload and param.default is inspect.Parameter.empty:
            raise TypeError(f"Missing required argument: {name}")

        value = payload.get(name, param.default)

        if param.annotation is not inspect._empty:
            adapter = TypeAdapter(param.annotation)
            value = adapter.validate_python(value)

        bound_args[name] = value

    if inspect.iscoroutinefunction(callback):
        return await callback(**bound_args)
    return callback(**bound_args)


def issubclass_safe(obj: Any, cls: type) -> bool:
    try:
        return inspect.isclass(obj) and issubclass(obj, cls)
    except Exception:
        return False



"""ATTP Client Library

A Python client library for the ATTP (Agent Transfer Transfer Protocol) framework.
"""

# Main client class
from .client import ATTPClient

# Core API classes
from .catalog import AttpCatalog
from .inference import AttpInferenceAPI
from .router import AttpRouter
from .session import SessionDriver
from .tools import ToolsManager

# Exception classes
from .errors.attp_exception import AttpException
from .errors.correlated_rpc_exception import CorrelatedRPCException
from .errors.dead_session import DeadSessionError
from .errors.not_found import NotFoundError
from .errors.serialization_error import SerializationError
from .errors.unauthenticated_error import UnauthenticatedError

# Interface classes
from .interfaces.error import IErr
from .interfaces.route_mappings import IRouteMapping
from .interfaces.catalogs.catalog import ICatalogResponse
from .interfaces.catalogs.tools.envelope import IEnvelope
from .interfaces.handshake.auth import IAuth
from .interfaces.handshake.hello import IHello
from .interfaces.handshake.ready import IReady
from .interfaces.inference.message import IMessageResponse, IMessageDTOV2
from .interfaces.inference.tool import ToolV2
from .interfaces.inference.enums.message_data_type import MessageDataTypeEnum
from .interfaces.inference.enums.message_emergency_type import MessageEmergencyTypeEnum
from .interfaces.inference.enums.message_type import MessageTypeEnum

# Type definitions
from .types.route_mapping import AttpRouteMapping, RouteType

# Utility classes
from .utils.context_awaiter import ContextAwaiter
from .utils.route_mapper import resolve_route_by_id
from .utils.serializer import serialize, deserialize

# Misc classes
from .misc.fixed_basemodel import FixedBaseModel
from .misc.serializable import Serializable

# Constants
from .consts import ATTP_VERSION

__version__ = "0.1.0"
__all__ = [
    # Main client
    "ATTPClient",
    
    # Core API classes
    "AttpCatalog",
    "AttpInferenceAPI", 
    "AttpRouter",
    "SessionDriver",
    "ToolsManager",
    
    # Exceptions
    "AttpException",
    "CorrelatedRPCException",
    "DeadSessionError",
    "NotFoundError",
    "SerializationError",
    "UnauthenticatedError",
    
    # Interfaces
    "IErr",
    "IRouteMapping",
    "ICatalogResponse",
    "IEnvelope",
    "IAuth",
    "IHello",
    "IReady",
    "IMessageResponse",
    "IMessageDTOV2",
    "ToolV2",
    "MessageDataTypeEnum",
    "MessageEmergencyTypeEnum",
    "MessageTypeEnum",
    
    # Types
    "AttpRouteMapping",
    "RouteType",
    
    # Utils
    "ContextAwaiter",
    "resolve_route_by_id",
    "serialize",
    "deserialize",
    
    # Misc
    "FixedBaseModel",
    "Serializable",
    
    # Constants
    "ATTP_VERSION",
]


import asyncio
from typing import Any, Callable, Mapping


def trigger_callable(
    callable: Callable[..., Any], 
    args: tuple[Any] | None = None, 
    kwargs: Mapping[str, Any] | None = None
):
    asyncio.create_task(callable(*(args or tuple()), **(kwargs or {})))


from typing import Annotated, Any, Mapping
from pydantic import Field
from typing_extensions import Doc
from attp_client.misc.fixed_basemodel import FixedBaseModel


class IEnvelope(FixedBaseModel):
    """Uniform tool output"""
    type: Annotated[str, Doc("Envelop discriminator; For tools only 'tool-call'")] = Field("tool-call")
    catalog: Annotated[str, Doc("Tool catalog name where tool is registered and being called from")]
    tool_id: Annotated[str, Doc("Tool ID being called")]
    metadata: Annotated[Mapping[str, Any], Doc("AgentHub/trace metadata")] = Field(default_factory=dict)
    data: Annotated[Mapping[str, Any], Doc("Tool output JSON matching provided tool arg schema")]


import asyncio

from attp_core.rs_api import PyAttpMessage, AttpCommand

from attp_client.errors.attp_exception import AttpException
from attp_client.interfaces.error import IErr


class StatefulAckGate:
    def __init__(self) -> None:
        self.pendings: dict[bytes, asyncio.Queue] = {}
        self.pending_lock = asyncio.Lock()
    
    async def request_ack(self, correlation_id: bytes):
        async with self.pending_lock:
            queue = self.pendings.get(correlation_id)
            if queue is None:
                queue = asyncio.Queue()
                self.pendings[correlation_id] = queue
        
        return queue

    async def feed(self, message: PyAttpMessage) -> None:
        if not message.correlation_id:
            return

        async with self.pending_lock:
            queue = self.pendings.get(message.correlation_id)
            if queue is None:
                return

        queue.put_nowait(message)
    
    async def wait_for_ack(
        self, 
        correlation_id: bytes, 
        timeout: float,
        *, queue: asyncio.Queue[PyAttpMessage] | None = None
    ):
        if not queue:
            queue = self.pendings.get(correlation_id) or await self.request_ack(correlation_id)
        
        while True:
            message = await asyncio.wait_for(queue.get(), timeout=timeout)
            
            if message.command_type == AttpCommand.ERR:
                raise AttpException.from_ierr(
                    err=IErr.mps(message.payload) if message.payload else IErr(detail={"code": "ErrorWithoutPayload"}),
                    correlation_id=correlation_id
                )
            if message.command_type == AttpCommand.DEFER:
                continue
            
            if message.command_type == AttpCommand.ACK:
                return message
    
    async def stream_ack(
        self,
        correlation_id: bytes,
        timeout: float,
        *, queue: asyncio.Queue[PyAttpMessage] | None = None
    ):
        if not queue:
            queue = self.pendings.get(correlation_id) or await self.request_ack(correlation_id)

        while True:
            message = await asyncio.wait_for(queue.get(), timeout=timeout)
            
            if message.command_type == AttpCommand.ERR:
                raise AttpException.from_ierr(
                    err=IErr.mps(message.payload) if message.payload else IErr(detail={"code": "ErrorWithoutPayload"}),
                    correlation_id=correlation_id
                )
            if message.command_type == AttpCommand.STREAMBOS:
                continue
            
            if message.command_type == AttpCommand.CHUNK:
                yield message
            
            if message.command_type == AttpCommand.STREAMEOS:
                return
        
    async def complete_ack(self, correlation_id: bytes):
        """Call when response is returned and theres no need for the corr_id to be hanging on pending"""
        async with self.pending_lock:
            self.pendings.pop(correlation_id, None)



from typing import Any

from attp_core.rs_api import PyAttpMessage

from attp_client.catalog import AttpCatalog
from attp_client.interfaces.error import IErr
from attp_client.misc.fixed_basemodel import FixedBaseModel
from attp_client.misc.serializable import Serializable
from attp_client.session import SessionDriver
from attp_client.tools import ToolsManager
from attp_client.core.eventbus import AttpEventBus
from attp_client.utils import envelopizer


class ToolCallbacks:
    def __init__(
        self,
        *,
        session: SessionDriver,
        catalogs: list[AttpCatalog],
        tools: ToolsManager,
    ) -> None:
        self._session = session
        self._catalogs = catalogs
        self._tools = tools

    async def register(self, eventbus: AttpEventBus) -> None:
        await eventbus.subscribe("message", "tools:call", self.handle_tool_call)
        await eventbus.subscribe("message", "catalogs:tools:list", self.list_tools)

    async def handle_tool_call(self, message: PyAttpMessage):
        if not message.correlation_id:
            await self._session.send_error(IErr(
                detail={"message": "Correlation ID was missing in the message.", "code": "MissingCorrelationId"},
            ), route=message.route_id)
            return

        print("TOOL CALLBACK MESSAGE:", message.payload)
        try:
            envelope = envelopizer.envelopize(message)
        except ValueError as e:
            await self._session.send_error(IErr(
                detail={"message": str(e), "code": "InvalidPayload"},
            ), correlation_id=message.correlation_id, route=message.route_id)
            return

        catalog = next((c for c in self._catalogs if c.catalog_name == envelope.catalog), None)
        if not catalog:
            await self._session.send_error(IErr(
                detail={"message": f"Catalog with name {envelope.catalog} not found.", "code": "NotFoundError"},
            ), route=message.route_id, correlation_id=message.correlation_id)
            return

        response = await catalog.handle_callback(envelope)

        if isinstance(response, IErr):
            await self._session.send_error(response, route=message.route_id, correlation_id=message.correlation_id)
            return

        if not isinstance(response, FixedBaseModel) and not isinstance(response, Serializable):
            response = Serializable[Any](response)

        await self._session.respond(route=message.route_id, correlation_id=message.correlation_id, payload=response)

    async def list_tools(self, message: PyAttpMessage):
        if not message.payload:
            await self._session.send_error(IErr(
                detail={"message": "Payload was missing in the message.", "code": "MissingPayload"},
            ), route=message.route_id)
            return

        deserialized = Serializable[dict[str, Any]].mps(message.payload)

        if deserialized.data.get("catalog_name") is None:
            await self._session.send_error(IErr(
                detail={"message": "Catalog name was missing in the payload.", "code": "MissingCatalogName"},
            ), route=message.route_id, correlation_id=message.correlation_id)
            return

        catalog_name = deserialized.data["catalog_name"]
        tools = self._tools.get_tools(catalog_name=catalog_name)

        return Serializable[dict[str, Any]]({
            "tools": [tool.model_dump(mode="json") for tool in tools]
        })



from dataclasses import dataclass
from typing import Any, Literal, TypeAlias

# from core.attp.interfaces.handshake.mapping import IRouteMapping

RouteType: TypeAlias = Literal["event", "message", "err", "disconnect", "connect"]


@dataclass(frozen=True, unsafe_hash=True)
class AttpRouteMapping:
    pattern: str
    route_id: int
    route_type: RouteType
    callback: Any